/*
 * gpe-applauncher
 *
 * Application launcher for GPE Phone Edition.
 * 
 * (c) kernel concepts 2006
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Application launcher user interface.
 */
#include <stdlib.h>
#include <sys/types.h>
#include <dirent.h>
#include <string.h>
#include <libintl.h>
#include <time.h>
#include <fcntl.h>
#include <unistd.h>

#include <gtk/gtk.h>
#include <gdk/gdkx.h>

#include <gpe/spacing.h>
#include <gpe/infoprint.h>
#include <gpe/launch.h>
#include <gpe/desktop_file.h>
#include <gpe/pixmaps.h>
#include <gpe/gtkappnotification.h>

#include "homescreen.h"
#include "applications.h"
#include "settings.h"
#include "gpe-applauncher.h"
#include "dbus-if.h"

#define _(x) gettext (x)

/* Number of displayed columns in icon and list layout. */
#define COLUMNS_STACK   3
#define COLUMNS_SIDE    1

#define VOC_PID "/tmp/.voc.pid"
#define ICON_LOADING "loading.gif"
#define SUPER_FOLDER_BG "super_folder_bg.png"
#define FOLDER_BG "folder_bg.png"

/* Data type for application folder/tab encapsulation. */
typedef struct {
	gint pagenum;         /* pagenum of the notebook tab */
    gchar *name;          /* name / label */
    gchar *iconname;      /* icon filename */
    gchar *category;      /* matching category, maybe wants to be extended to multiple */
    GtkWidget *icon;      /* icon widget */
    GtkTreeModel *storage;/* associated application item data */
    GtkWidget *iconview;  /* widget used to display data */
} t_folder;

/* Application property fields for list store. */
typedef enum {
    APROP_NAME,
    APROP_COMMENT,
    APROP_EXEC,
    APROP_TERMINAL,
    APROP_STARTUP,
    APROP_SINGLE,
    APROP_ICONNAME,
    APROP_ICONDATA,
    APROP_TYPE,
    APROP_MAX
} t_app_property;


static GList *folderlist;
static t_folder *default_folder = NULL;

static GHashTable * notification_table = NULL;

/* to test whether a lock file can be acquired with Write permission
   return 0 means OK
   return 1 means another process is having lock on this file

   The Voc application will create a lock file when started, the file will 
   exist in the system when Voc is running until Voc exits. Applauncher tests
   the existence of this file to determine whether Voc is running (return 1)
   or not (return 0)
*/
static int 
test_lock(const gchar * lock_file)
{
  int fd, ret;
  struct flock lock_it;

  fd = open (lock_file, O_WRONLY);
  
  lock_it.l_type = F_WRLCK;
  lock_it.l_whence = SEEK_SET;
  lock_it.l_start = 0;
  lock_it.l_len = 0;

  if (fcntl (fd, F_GETLK, &lock_it) < 0)
      ret = 0;
  else if (lock_it.l_type == F_UNLCK)
      ret = 0;
  else 
      ret = 1;
  
  close (fd);
  return ret;
}

static gint get_pagenum (const char * name)
{
    GList *iter;
    gint pagenum;
    for (iter = folderlist; iter; iter = iter->next)
    {
      t_folder *folder = iter->data;
      if (strcmp(name, folder->name) == 0)
      {
        pagenum = folder->pagenum;
        break;
      }
    }
    return pagenum;
}

static GtkWidget * retrieve_notebook (void)
{
    GList *iter = folderlist;
    if (!iter)
      return NULL;
    t_folder *folder = iter->data;
    if (folder)
    {
         GtkWidget *parent = folder->iconview;
         while (!GTK_IS_NOTEBOOK (parent) && parent != NULL)
         {
               parent = gtk_widget_get_parent (parent);
         }
         if (GTK_IS_NOTEBOOK (parent) && parent != NULL)
           return parent;
    }
    return NULL;
}

static void
render_bg (GtkWidget *widget, const char* filename)
{
  GError *error = NULL;
  GdkPixbuf *gdkpixbuf;
  GdkPixmap *bgPixmap;
  GdkBitmap *bgMask;
  gdkpixbuf = gdk_pixbuf_new_from_file (filename, &error);
  if (error != NULL)
  {
          g_print ("unable to find background file: %s\n", error->message);
          g_error_free (error);
  }
  gdk_pixbuf_render_pixmap_and_mask (gdkpixbuf, &bgPixmap, &bgMask, 1);
  if (bgPixmap == NULL)
  {
          g_print ("no pixmap created!\n");
  }
   
   GtkStyle   *style;
   
   style = gtk_style_copy (widget->style);
   
   if (style->bg_pixmap[0])
     g_object_unref (style->bg_pixmap[0]);
   style->bg_pixmap[0] = g_object_ref (bgPixmap);
   gtk_widget_set_style (widget, style);
   g_object_unref (style);
 
//  GtkStyle * gtkStyle;
//  gtkStyle = gtk_rc_get_style (widget);
//  gtkStyle->bg_pixmap[0] = bgPixmap;
//  gtk_style_set_background (gtkStyle, widget->window, GTK_STATE_NORMAL);
}

/* Update stored application pixmaps in order to support setting changes. */
static gboolean
update_item_pixmap (GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter,
                    gpointer data)
{
  GdkPixbuf *pbuf = NULL;
  gchar *iconname;
  gint size = (gint)data;
    
  gtk_tree_model_get (model, iter, 
                      APROP_ICONDATA, &pbuf, APROP_ICONNAME, &iconname, -1);
  if (pbuf)
    g_object_unref (pbuf);
  
  pbuf = load_icon_scaled (iconname, size);
  
  if (!pbuf)
    {
      pbuf = gpe_find_icon ("applauncher");
      pbuf = gdk_pixbuf_scale_simple (pbuf, applauncher_settings.item_size,
                                      applauncher_settings.item_size, 
                                      GDK_INTERP_BILINEAR);
    }
  
  gtk_list_store_set (GTK_LIST_STORE (model), iter, 
                      APROP_ICONDATA, pbuf, -1);
    
  if (pbuf)
    g_object_unref (pbuf);
 
  return FALSE;
}

/* Handle theme changes. */
void 
applications_update_theme (const gchar *newtheme)
{
  GList *iter;
  
  if (newtheme == NULL)
    return;
    
  for (iter = folderlist; iter; iter = iter->next)
    {
      t_folder *folder = iter->data;
      GdkPixbuf *pbuf = NULL, *iconbuf = NULL;
      gchar *str = NULL;
      gchar *gpe_path;

      gpe_path = getenv ("GPE_PATH");
      if (gpe_path)
        {          
          str = g_build_filename(gpe_path, "share/gpe/pixmaps/default", folder->iconname, NULL);
          pbuf = gdk_pixbuf_new_from_file (str, NULL);
        }
      if (pbuf == NULL)
        { 
          if (str)
              g_free (str);		  
          str = g_build_filename (DEFAULT_THEMEDIR, newtheme, THEMEFOLDER, folder->iconname, NULL);
          pbuf = gdk_pixbuf_new_from_file (str, NULL);
          if (pbuf == NULL)
            {
              g_free (str);	
              str = g_build_filename (LOCAL_THEMEDIR, newtheme, THEMEFOLDER, folder->iconname, NULL);
              pbuf = gdk_pixbuf_new_from_file (str, NULL);
              if (pbuf == NULL)
                {
                  g_free (str);
                  str = g_build_filename (GPE_PIXMAPS_DIR, folder->iconname, NULL);
                  pbuf = gdk_pixbuf_new_from_file (str, NULL);
                }
            }
        }
    
    g_free (str);
    if (pbuf == NULL)
        continue;
      
    iconbuf = gdk_pixbuf_scale_simple (pbuf, TAB_ICON_SIZE, 
                                     TAB_ICON_SIZE, GDK_INTERP_BILINEAR);
    gtk_image_set_from_pixbuf (GTK_IMAGE (folder->icon), iconbuf);
    g_object_unref (pbuf);
    g_object_unref (iconbuf);
  }
  /* set background image for every folder */
  for (iter = folderlist; iter; iter = iter->next)
  {
      t_folder *folder = iter->data;
      gchar *str = NULL;
      gchar *gpe_path;
      gboolean found = FALSE;

      gpe_path = getenv("GPE_PATH");
      if (gpe_path)
      {
          str = g_build_filename(gpe_path, "share/themes/", newtheme, THEMEFOLDER, FOLDER_BG, NULL);
      }
      found = access (str, R_OK) ? FALSE : TRUE;
      if (!found)
      {
          if (str) g_free (str);
          str = g_strdup_printf ("%s/%s/%s/%s", DEFAULT_THEMEDIR, newtheme,
                                 THEMEFOLDER, FOLDER_BG);
          found = access (str, R_OK) ? FALSE : TRUE;
      }
      if (!found)
      {
          g_free (str);
          str = g_strdup_printf ("%s/%s/%s/%s", LOCAL_THEMEDIR, newtheme,
                                   THEMEFOLDER, FOLDER_BG);
          found = access (str, R_OK) ? FALSE : TRUE;
      }

      GtkWidget *parent = folder->iconview;
      while (!GTK_IS_WINDOW (parent) && parent != NULL)
      {
           parent = gtk_widget_get_parent (parent);
      }
      if (GTK_IS_WINDOW (parent) && parent != NULL)
      {
        if (found)
        {
          g_print ("folder background file: %s\n", str);
          if (!GTK_WIDGET_NO_WINDOW(folder->iconview))
            g_print ("iconview has its own window\n");
          render_bg (folder->iconview, str);
          //set_background_pixmap (str, folder->iconview);
        }
        else
          set_background_pixmap (NULL, folder->iconview);
      }
      if (str)
          g_free (str);
  }
  
}

/* Change background image of application area - currently unsupported. */
void 
applications_update_background (GtkWidget *window)
{
  GList *iter;
    
  for (iter = folderlist; iter; iter = iter->next)
    {
      t_folder *folder = iter->data;
                  
      gdk_window_set_back_pixmap (folder->iconview->window,  
                                  window->style->bg_pixmap[GTK_STATE_NORMAL],
                                  FALSE);
      gtk_widget_queue_draw (folder->iconview);
    }
}

/* Change the size of the application icons. */
void
applications_update_itemsize (gint width)
{
  GList *iter;

  gint colwidth;

  colwidth = (gdk_screen_width() - 20) / (applauncher_settings.item_layout ? COLUMNS_SIDE : COLUMNS_STACK);
	
  if (applauncher_settings.item_layout == LAYOUT_SIDE)
      width /= 2;
    
  for (iter = folderlist; iter; iter = iter->next)
    {
      t_folder *folder = iter->data;
      gtk_icon_view_set_item_width (GTK_ICON_VIEW(folder->iconview), colwidth);
      gtk_tree_model_foreach (GTK_TREE_MODEL(folder->storage), update_item_pixmap, (gpointer) width);
    }
}

/* Change the layout of the application displays. */
void
applications_update_item_layout (gint type)
{
  GList *iter;
  GtkOrientation or; 
    
   switch (type)
     {
        case LAYOUT_STACK:
            or = GTK_ORIENTATION_VERTICAL;
        break;
        case LAYOUT_SIDE:
            or = GTK_ORIENTATION_HORIZONTAL;
        break;
        default:
            or = GTK_ORIENTATION_VERTICAL;
     }

   for (iter = folderlist; iter; iter = iter->next)
      {
        t_folder *folder = iter->data;
        gtk_icon_view_set_orientation (GTK_ICON_VIEW(folder->iconview), or);
        gtk_icon_view_set_columns (GTK_ICON_VIEW (folder->iconview), type ? COLUMNS_SIDE : COLUMNS_STACK);
      }
   applications_update_itemsize (applauncher_settings.item_size);
}

static gboolean 
launch_notification_timeout_cb (gpointer callee)
{
  GtkWidget * nc_window = NULL;
            
  nc_window = (GtkWidget *)g_hash_table_lookup(notification_table, callee);
  if (nc_window)
    {
      gtk_widget_destroy(nc_window);
      g_hash_table_remove(notification_table, callee);
    }

  g_free(callee);

  return FALSE;
}

/* Callback used to track the startup status of applications. */
static void
launch_status_callback (enum launch_status status, void *data)
{
  gchar *callee = NULL;
    
  if (data)
    {
      gchar ** vec = g_strsplit_set((gchar*)data, "!", -1);
		
      if (vec[0])
        {
          callee = g_strdup(vec[0]);
          g_print("%s(): callee '%s'", __FUNCTION__,
							callee ? callee : "null");
        }
		
        g_strfreev(vec);
    }
   
  switch (status)
    {
      case LAUNCH_STARTING:
#ifdef DEBUG
        g_message ("Launch starting...");
#endif        
        if (callee)
          {
            GtkWidget *nc_window = NULL;
            gchar *file;
            
            /* find loading icon */
            if (getenv("GPE_PATH"))
              {
                file = g_build_filename(getenv("GPE_PATH"), "share/themes/", 
                    applauncher_settings.themename ? applauncher_settings.themename : "default",
                    THEMEFOLDER, ICON_LOADING, NULL);
              }
            else
              {
                file = g_build_filename (DEFAULT_THEMEDIR, 
                                        applauncher_settings.themename ? applauncher_settings.themename : "default", 
    	                                THEMEFOLDER, ICON_LOADING, NULL);
              }
            if (!g_file_test (file, G_FILE_TEST_EXISTS))
              {
                g_free (file);
                file = g_build_filename (LOCAL_THEMEDIR, 
                                         applauncher_settings.themename ? applauncher_settings.themename : "default",
                                         THEMEFOLDER, ICON_LOADING, NULL);
              }
             
            g_print ("\nloading icon: %s\n", file);
                
            nc_window = gtk_app_notification_new();
            gtk_widget_set_size_request(GTK_WIDGET(nc_window), -1, 70);
            gtk_app_notification_set_text(GTK_APP_NOTIFICATION(nc_window), 
                                          _("Launching..."));
    
            gtk_app_notification_set_image_from_file(GTK_APP_NOTIFICATION(nc_window), 
                                                     file);
            g_free (file);
            gtk_app_notification_hide_progressbar(GTK_APP_NOTIFICATION(nc_window));
            
            gtk_window_set_position(GTK_WINDOW(nc_window), GTK_WIN_POS_CENTER_ALWAYS);

            g_hash_table_replace(notification_table, (gpointer)callee, (gpointer)nc_window);
                                                     
            g_timeout_add(4 * 1000, launch_notification_timeout_cb, g_strdup(callee));
                                                     
            gtk_widget_show_all(nc_window);
          }
        
        break;
      
    case LAUNCH_COMPLETE:
#ifdef DEBUG
        g_message ("Launch completed!");
#endif        
        if (data)
          {
            send_message (data);		
            g_free (data);
          }
        if (callee)
          {
            GtkWidget *nc_window = NULL;
            
            nc_window = (GtkWidget *)g_hash_table_lookup(notification_table, (gpointer)callee);
            if (nc_window)
              {
                gtk_widget_destroy(nc_window);
                g_hash_table_remove(notification_table, (gpointer)callee);
              }
          }
        break;
    case LAUNCH_FAILED:      		
#ifdef DEBUG
        g_message ("Launch failed!");
#endif
        if (applauncher_settings.themename)
          applications_update_theme (applauncher_settings.themename);
        g_free (data);
        break;
    }
}

gboolean 
launch_application (const char *name, const char *executable, const char *argv,
                    gboolean single_instance, gboolean startup_notify,
					const char *caller)
{
  Window w;
  gchar *s;
  gboolean ret = FALSE;
    
  if (!name || !executable || !argv || !caller)
      return FALSE;
  
  gchar *exec = g_strdup (executable);
    
  if (!startup_notify)
      single_instance = FALSE;
    
  s = strstr (exec, "%U");
  if (!s)
      s = strstr (exec, "%f");
  if (s)
      *s = 0;
                      
  gchar *exec_argv = g_strconcat (exec, " ", argv, NULL);
  gchar *str_send = g_strdup_printf ("%s!%s!%s", name, argv, caller);
  
#ifdef DEBUG
  g_message ("name = '%s'", name);
  g_message ("exec = '%s'", exec);
  g_message ("exec_argv = '%s'", exec_argv);
  g_message ("caller = '%s'", caller);
#endif
  
  if (!single_instance)
    {
      /* run another instance */  											  
      ret = gpe_launch_program_with_callback (GDK_DISPLAY(), exec_argv, name, 
                                              startup_notify,
                                              launch_status_callback, str_send);
    }
  else
    { 
      /*Voc will place a file in /tmp to advertise its existence */
      if (strcmp (name , "Voc") == 0) 
        {
          if (test_lock (VOC_PID))
            {
              send_message (str_send);	
              g_free (str_send);
              ret = TRUE;
            }
          else
            {			
              ret = gpe_launch_program_with_callback (GDK_DISPLAY(), exec_argv, 
                                                      name, TRUE,
                                                      launch_status_callback, str_send);
            }
        }
      else /* the running-or-not determination of the other applications are 
           conducted by checking with its window */
        {
            w = gpe_launch_get_window_for_binary (GDK_DISPLAY(), exec);  		  
            if (w)
              {					  
                /* tell the caller that startup is successful */  
#ifdef DEBUG
                g_message ("activate it");  
#endif
                if (strcmp (argv, "") == 0)
                    gpe_launch_activate_window (GDK_DISPLAY(), w);
                else
                  {
                    send_message (str_send);	
                    g_free (str_send);
                  }
                ret = TRUE;						
              }
            else
              {
                if (!gpe_launch_startup_is_pending (GDK_DISPLAY(), exec))
                  {
                    /* Actually run the program */
#ifdef DEBUG
                    g_message ("launch it");
#endif
                    ret = gpe_launch_program_with_callback (GDK_DISPLAY(), 
                                            exec_argv, name, TRUE,
                                            launch_status_callback, str_send);
                  }
              } 
        }
    }
  g_free (exec_argv);
  g_free (exec);
  return ret;
}

int 
applications_start_service (const gchar *service_name, const gchar *method, 
							const gchar* caller_service_name)
{
  GList *iter;
  gchar *name = NULL;
  gchar *exec = NULL;
  gboolean single_instance = TRUE;
  gboolean startup_notify = TRUE;
  gboolean ret;	
  gint out = 0;

  if (caller_service_name != NULL)
    g_message ("caller is %s", caller_service_name);

  for (iter = folderlist; iter; iter = iter->next)
    {
      if (out == 1)
          break;
      
      t_folder *folder = iter->data;
      GtkTreeIter *tree_iter;                
      GtkTreeIter first_iter;          
              
      if (gtk_tree_model_get_iter_first (folder->storage, &first_iter))
        tree_iter = &first_iter;
      else /* an empty folder */
        continue;
      do
        {
          gtk_tree_model_get (folder->storage, tree_iter, APROP_NAME, &name, -1);			       
          if (strcmp (service_name, name) == 0)
            {
              gtk_tree_model_get (folder->storage, tree_iter, APROP_EXEC, &exec, 
                                  APROP_SINGLE, &single_instance,					   						 
                                  APROP_STARTUP, &startup_notify, -1);
              out = 1;
              break;
            }   
          if (name) {g_free (name); name = NULL;}
        } 
      while (gtk_tree_model_iter_next (folder->storage, tree_iter));
	}
	
  g_message ("%s is %s\n", 
						 (service_name ? service_name : "(null)"),
						 (exec ? exec : "(null)"));

  ret = launch_application (name, exec, method, single_instance, 
                            startup_notify, caller_service_name);

  if (exec) 
      g_free (exec); 
  if (name) 
      g_free (name); 
	
  return (ret ? 0 : 1);
}
	
/* Start an application honoring single instance applications and notification. */
static void
applications_activate (GtkIconView *iconview, GtkTreePath *path, gpointer user_data)
{
  t_folder *folder = (t_folder*)user_data;
  GtkTreeIter iter;
  gchar *exec;
  gchar *name;
  gchar *type;
  gboolean single_instance = TRUE;
  gboolean startup_notify;
  gchar *text;
  Display *dpy = GDK_DISPLAY();
    
  gtk_tree_model_get_iter (folder->storage, &iter, path);
  gtk_tree_model_get(folder->storage, &iter, APROP_EXEC, &exec, 
                                             APROP_SINGLE, &single_instance,
                                             APROP_STARTUP, &startup_notify,
                                             APROP_NAME, &name,
                                             APROP_TYPE, &type, -1);
  if (strcmp (type, "Directory") == 0)
  {
       GList *iter;
       gint pagenum;
    for (iter = folderlist; iter; iter = iter->next)
    {
      t_folder *folder = iter->data;
      if (strcmp(name, folder->name) == 0)
      {
        pagenum = folder->pagenum;
        break;
      }
    }
    GtkWidget *parent = folder->iconview;
    while (!GTK_IS_NOTEBOOK (parent) && parent != NULL)
    {
        parent = gtk_widget_get_parent (parent);
    }
    if (GTK_IS_NOTEBOOK (parent) && parent != NULL)
      gtk_notebook_set_current_page (GTK_NOTEBOOK (parent), pagenum);
  }
  else
  {  
    if (name == NULL) 
      name = g_strdup (exec);

    text = g_strdup_printf ("%s %s", _("Starting"), name);
    gpe_popup_infoprint (dpy, text);
    g_free (text);
  
    launch_application (name, exec, "", single_instance, startup_notify, 
                      SERVICE_APPLAUNCHER);
  }	
  if (exec) g_free (exec); 
  if (name) g_free (name); 
  if (type) g_free (type);
  
  return;
}

/* Create a new application folder widget from its desktop description. */
static t_folder*
new_folder_from_desktop (GnomeDesktopFile *g)
{
  t_folder *result = g_malloc0 (sizeof(t_folder));
  gchar *str;
  GdkPixbuf *pbuf, *iconbuf;
  gchar * gpe_path = NULL;
	
  gnome_desktop_file_get_string (g, NULL, "Name", &str);
  result->name = str;
  gnome_desktop_file_get_string (g, NULL, "Icon", &str);
  result->iconname = str;
  gnome_desktop_file_get_string (g, NULL, "Match", &str);
  result->category = str;  
  
  gpe_path = getenv ("GPE_PATH");
  if (gpe_path)
    {
      str = g_build_filename( gpe_path, "share/gpe/pixmaps/default", 
                              result->iconname, NULL);
    }
  else
    {
      str = g_strdup_printf ("%s/%s", GPE_PIXMAPS_DIR, result->iconname);
    }
  
  pbuf = gdk_pixbuf_new_from_file (str, NULL);
  if (!pbuf)
    {
      pbuf = gpe_find_icon ("applauncher");
      /* just a copy to the correct size */
      pbuf = gdk_pixbuf_scale_simple (pbuf, TAB_ICON_SIZE, TAB_ICON_SIZE, 
                                      GDK_INTERP_BILINEAR);	
    }
  
  iconbuf = gdk_pixbuf_scale_simple (pbuf, TAB_ICON_SIZE, 
                                     TAB_ICON_SIZE, GDK_INTERP_BILINEAR);
  result->icon = gtk_image_new_from_pixbuf (iconbuf);
  gtk_widget_set_usize (result->icon, (gdk_screen_width() - 42) / 4, -1);
    
  g_object_unref (pbuf);
  g_object_unref (iconbuf);
  g_free (str);
 
  /* is this our default folder? */
  if (!strcmp("Other", result->category))
      default_folder = result;
    
  return result;
}

/* Create a new allpication data set from its desktop description and add this
 * to the list store. */
static void
new_application_from_desktop (GnomeDesktopFile *g, GtkTreeModel *store)
{
  GtkTreeIter iter;
  gchar *name, *icon, *exec, *type;
  gboolean terminal;
  gboolean startup;
  gboolean single = TRUE;
  gboolean bv;
  GdkPixbuf *ibuf;
    
  gnome_desktop_file_get_string (g, NULL, "Name", &name);
  gnome_desktop_file_get_string (g, NULL, "Icon", &icon);
  gnome_desktop_file_get_string (g, NULL, "Exec", &exec);
  gnome_desktop_file_get_string (g, NULL, "Type", &type);
  if (gnome_desktop_file_get_boolean (g, NULL, "Terminal", &terminal) == FALSE)
      terminal = FALSE;
  if (gnome_desktop_file_get_boolean (g, NULL, "StartupNotify", &startup) == FALSE)
      startup = TRUE;
  if (gnome_desktop_file_get_boolean (g, NULL, "SingleInstance", &bv) && (bv == FALSE))
      single = FALSE;
  if (gnome_desktop_file_get_boolean (g, NULL, "X-SingleInstance", &bv) && (bv == FALSE))
      single = FALSE;

  ibuf = load_icon_scaled (icon, applauncher_settings.item_size);
  
  if (!ibuf)
    {
      ibuf = gpe_find_icon ("applauncher");
      ibuf = gdk_pixbuf_scale_simple (ibuf, applauncher_settings.item_size,
                                      applauncher_settings.item_size, GDK_INTERP_BILINEAR);	
    }
  
  if (ibuf)
    {
      gtk_list_store_append (GTK_LIST_STORE (store), &iter);
      gtk_list_store_set (GTK_LIST_STORE (store), &iter, APROP_NAME, name,
                          APROP_ICONDATA, ibuf, APROP_ICONNAME, icon, 
                          APROP_EXEC, exec, 
                          APROP_TERMINAL, terminal, APROP_STARTUP, startup, 
                          APROP_SINGLE, single, APROP_TYPE, type, -1);
      g_object_unref (ibuf);
    }
  g_free (name);
  g_free (icon);
  g_free (type);
  
}

/* Load all available desktop descriptions of folders. */
static GList*
load_folder_list (const gchar *path)
{
  DIR *dir;
  struct dirent *entry;
  GList *result = NULL;
  t_folder *new_folder;

  dir = opendir (path);
  if (dir)
    {
      while ((entry = readdir (dir)))
      {
	  char *temp;
	  GnomeDesktopFile *p;
	  GError *err = NULL;
	  
	  if (entry->d_name[0] == '.')
	    continue;
	
      temp = g_strdup_printf ("%s/%s", path, entry->d_name);
	  p = gnome_desktop_file_load (temp, &err);
	  if (p)
	    {
	      gchar *type;
	      gnome_desktop_file_get_string (p, NULL, "Type", &type);

	      if (type == NULL || strcmp (type, "Directory"))
              gnome_desktop_file_free (p);
	      else
            {
              new_folder = new_folder_from_desktop (p);
              result = g_list_append (result, new_folder);
              gnome_desktop_file_free (p);
            }

	      if (type)
             g_free (type);
        }
	  else
	    fprintf (stderr, "Couldn't load \"%s\": %s\n", temp, err->message);

	  if (err)
	    g_error_free (err);

	  g_free (temp);
    }
    closedir (dir);
  }
  return result;
}

/* Load all available application descriptions. */
void
load_application_list (const gchar *path)
{
  DIR *dir;
  struct dirent *entry;

  dir = opendir (path);
  if (dir)
    {
      while ((entry = readdir (dir)))
        {
          char *temp;	  
      
          if (entry->d_name[0] == '.')
              continue;
    
          temp = g_strdup_printf ("%s/%s", path, entry->d_name);
#ifdef DEBUG
          g_message ("load app: %s", entry->d_name);
#endif          
          add_application (temp);	 
     
          g_free (temp);
        }
      closedir (dir);	 
    }
}

/* Add a new application description to a folder */
void
add_application (const gchar *filename)
{
  GnomeDesktopFile *p;
  GError *err = NULL;
  p = gnome_desktop_file_load (filename, &err);
  if (p)
    {
      gchar *type;
      gnome_desktop_file_get_string (p, NULL, "Type", &type);

      if (type == NULL)
          gnome_desktop_file_free (p);
      else
        {
          GList *iter;
          gchar *cat;              
          gnome_desktop_file_get_string (p, NULL, "Categories", &cat);
          for (iter = folderlist; iter; iter = iter->next)
            {
              t_folder *folder = iter->data;
              if (cat && strstr(cat, folder->category))
                {
                  new_application_from_desktop (p, folder->storage);
                  //break;
                }
            }
          if ((iter == NULL) && (default_folder != NULL)) /* no folder found */
              new_application_from_desktop (p, default_folder->storage);
                  
          g_free (cat);
          gnome_desktop_file_free (p);
        }
      if (type)
          g_free (type);
    }
  else
     fprintf (stderr, "Couldn't load \"%s\": %s\n", filename, err->message);

  if (err)
     g_error_free (err);
}

/* Remove a new application description from a folder*/
void
remove_application (const gchar *filename)
{
  GnomeDesktopFile *p;
  GError *err = NULL;
  p = gnome_desktop_file_load (filename, &err);
  if (p)
    {
      gchar *type;
      gnome_desktop_file_get_string (p, NULL, "Type", &type);

      if (type == NULL || strcmp (type, "Application"))
          gnome_desktop_file_free (p);
      else
        {
          GList *iter;
          gchar *name;              
          gchar *cat;              
          gnome_desktop_file_get_string (p, NULL, "Categories", &cat); 
          gnome_desktop_file_get_string (p, NULL, "Name", &name);
          for (iter = folderlist; iter; iter = iter->next)
            {
              t_folder *folder = iter->data;
              if (cat && strstr(cat, folder->category))
                {
                  GtkTreeIter tree_iter;                
                  gchar *tree_iter_name;
                  gboolean end_reached;
 
                  end_reached = gtk_tree_model_get_iter_first (folder->storage, &tree_iter);
                  while (!end_reached)
                    {
                      gtk_tree_model_get(folder->storage, &tree_iter, APROP_NAME, &tree_iter_name, -1);
                      if (strcmp (name, tree_iter_name) == 0)
                        {
                          gtk_list_store_remove (GTK_LIST_STORE(folder->storage), &tree_iter);
                          g_free (tree_iter_name);
                          break; 
                        }
                      g_free (tree_iter_name);
                      end_reached = gtk_tree_model_iter_next (folder->storage, &tree_iter);
                    }           
                  break;
                }
          }          
          g_free (cat);        
          g_free (name);
          gnome_desktop_file_free (p);
       }
       if (type)
           g_free (type);
  }
  else
     fprintf (stderr, "Couldn't find \"%s\": %s\n", filename, err->message);

  if (err)
     g_error_free (err);
}


static void 
selection_iconview (GtkIconView *iconview, gpointer user_data)
{
  GtkTreeIter iter, former_iter;
  GtkTreePath *former_path = NULL; 
  gchar *icon, *former_icon;
  GdkPixbuf *ibuf, *former_ibuf;
  gchar *path_str, *former_path_str;
   
  GtkTreeModel *store = gtk_icon_view_get_model (iconview);
  GList *list = gtk_icon_view_get_selected_items (iconview);
    
  /* reset the formerly selected item's icon to normal size */
  former_path_str = g_object_get_data (G_OBJECT (iconview), "former_selected");
  if (former_path_str)
      former_path = gtk_tree_path_new_from_string (former_path_str);
  if (former_path)
    {
      gtk_tree_model_get_iter (store, &former_iter, former_path);
      gtk_tree_model_get (store, &former_iter, APROP_ICONNAME, &former_icon, -1);
      former_ibuf = load_icon_scaled (former_icon, applauncher_settings.item_size);  
      if (!former_ibuf)
        {
          former_ibuf = gpe_find_icon ("applauncher");
          former_ibuf = gdk_pixbuf_scale_simple (former_ibuf, applauncher_settings.item_size,
                                  applauncher_settings.item_size, GDK_INTERP_BILINEAR);	
        }
      if (former_ibuf)
        {
          gtk_list_store_set (GTK_LIST_STORE (store), &former_iter, APROP_ICONDATA, former_ibuf, -1);
          g_object_unref (former_ibuf);
        }
      g_free (former_path);
      g_free (former_icon);
   }
      
  if (g_list_length (list) > 0)
    {
      /* today is the yesterday of tomorrow */
      path_str = gtk_tree_path_to_string ((GtkTreePath*) list->data);
      g_object_set_data (G_OBJECT (iconview), "former_selected", path_str);		
  
      /* the selected item will have a slightly enlarged icon */	
      gtk_tree_model_get_iter (store, &iter, (GtkTreePath*) list->data);
      gtk_tree_model_get (store, &iter, APROP_ICONNAME, &icon, -1);	
      ibuf = load_icon_scaled (icon, applauncher_settings.item_size + 6);  
      if (!ibuf)
        {
          ibuf = gpe_find_icon ("applauncher");
          ibuf = gdk_pixbuf_scale_simple (ibuf, applauncher_settings.item_size + 6,
                                  applauncher_settings.item_size + 6, GDK_INTERP_BILINEAR);	
        }
      if (ibuf)
        {
          gtk_list_store_set (GTK_LIST_STORE (store), &iter, APROP_ICONDATA, ibuf, -1);
          g_object_unref (ibuf);
        }
      g_free (icon);
  }
    
  g_list_foreach (list, (GFunc)(gtk_tree_path_free), NULL);

  g_list_free (list);   
}

#ifdef GRIDLAYOUT
void set_super_folder_visible (void)
{
       GtkWidget *notebook = retrieve_notebook ();
       if (!notebook)
         return;
       gint pagenum = get_pagenum ("Super");
       gtk_notebook_set_current_page (GTK_NOTEBOOK (notebook), pagenum);
}

gboolean is_super_folder_visible (void)
{
    GtkWidget *notebook = retrieve_notebook ();
    if (!notebook)
      return FALSE;
    gint pagenum = get_pagenum ("Super");
    gint cur_pagenum = gtk_notebook_get_current_page (GTK_NOTEBOOK (notebook));
    if (cur_pagenum == pagenum)
      return TRUE;
    else
      return FALSE;
}
#endif

/* Create a new application tab from a given folder description. */
static void
new_application_tab (GtkWidget *notebook, t_folder *folder)
{
    GtkWidget *filebox, *sw;
    GtkTreeModel *store;
    
    store = GTK_TREE_MODEL (gtk_list_store_new (APROP_MAX, 
                                G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, 
                                G_TYPE_BOOLEAN, G_TYPE_BOOLEAN, 
                                G_TYPE_BOOLEAN, G_TYPE_STRING, GDK_TYPE_PIXBUF, G_TYPE_STRING));
    
    sw = gtk_scrolled_window_new (NULL, NULL);
    gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (sw), GTK_POLICY_NEVER, 
                                    GTK_POLICY_AUTOMATIC);
	
    filebox = gtk_icon_view_new_with_model (store);
    gtk_icon_view_set_columns (GTK_ICON_VIEW (filebox), COLUMNS_STACK);
    gtk_icon_view_set_spacing (GTK_ICON_VIEW (filebox), 0);
    gtk_icon_view_set_column_spacing (GTK_ICON_VIEW (filebox), gpe_get_boxspacing());
    gtk_icon_view_set_row_spacing (GTK_ICON_VIEW (filebox), gpe_get_boxspacing());
    gtk_icon_view_set_item_width (GTK_ICON_VIEW (filebox), -1);
	gtk_icon_view_set_markup_column (GTK_ICON_VIEW (filebox), APROP_NAME);
    gtk_icon_view_set_pixbuf_column (GTK_ICON_VIEW (filebox), APROP_ICONDATA);
    folder->iconview = filebox;	
	
    g_signal_connect (G_OBJECT(filebox), "item-activated", G_CALLBACK (applications_activate), folder);
	g_signal_connect (G_OBJECT(filebox), "selection-changed", G_CALLBACK (selection_iconview), NULL);
    
    folder->storage = gtk_icon_view_get_model (GTK_ICON_VIEW(filebox));
    
    /* packing */
    gtk_container_add (GTK_CONTAINER(sw), filebox);
    folder->pagenum = gtk_notebook_append_page (GTK_NOTEBOOK (notebook), sw, folder->icon);
}

/* Main function creating and setting up the application launcher user interface. */
void
applications_create (GtkWidget *notebook)
{
  GList *iter;
  gchar *gpe_path;

  gpe_path = getenv ("GPE_PATH");
  if (gpe_path)
    {
      gchar *str = g_build_filename(gpe_path, "share/gpe/vfolders", NULL);
      folderlist = load_folder_list(str);
      g_free(str);
    }
  else
    {
      folderlist = load_folder_list (PREFIX "/share/gpe/vfolders");
    }
    
   for (iter = folderlist; iter; iter = iter->next)
     {
       t_folder *folder = (t_folder*) iter->data;         
       new_application_tab (notebook, folder);
     }
	 
   if (gpe_path)
     {
       gchar * str = g_build_filename(gpe_path, "share/applications", NULL);
       load_application_list (str);
       g_free(str);
     }
   else
     { 
       load_application_list (PREFIX "/share/applications");
     }
   	 
   for (iter = folderlist; iter; iter = iter->next)
     {
       t_folder *folder = iter->data;
       GtkTreePath *treepath = gtk_tree_path_new_from_string ("0");
	   gtk_icon_view_select_path (GTK_ICON_VIEW (folder->iconview), treepath);
	   gtk_tree_path_free (treepath);	 
     }	   
  applications_update_item_layout (LAYOUT_STACK);
  /* initialize startup notification */
  gpe_launch_install_filter ();
  gpe_launch_monitor_display (GDK_DISPLAY());
  
  notification_table = g_hash_table_new_full(g_str_hash, g_str_equal,
                                             g_free, NULL);
}
