/*
 * gpe-applauncher
 *
 * Application launcher for GPE Phone Edition.
 * 
 * (c) kernel concepts 2006
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * 
 *  Xsettings configuration and settings dispatcher.
 */

#include <stdio.h>
#include <string.h>
#include <assert.h>

#include <gdk/gdkx.h>
#include <libintl.h>
#include <gtk/gtk.h>
#include <xsettings-client.h>

#include "settings.h"
#include "homescreen.h"
#include "applications.h"
#include "softkeys.h"
#include "gpe-applauncher.h"

#define _(x) gettext(x)

static XSettingsClient *client;

/* Some Xsettings namespace definitions */
#define KEY_BASE  "GPE/applauncher/"
#define KEY_THEME "Net/"

enum image_type
  {
    CENTERED,
    TILED,
    SCALED
  };

/* internal setting status and defaults */
t_settings applauncher_settings = { NULL, NULL, 48, NULL, NULL, 0 };

/* Theme change event handler. */
static void
update_theme (const gchar* newtheme)
{

  if (!applauncher_settings.themename || strcmp (applauncher_settings.themename, newtheme))
    {
      if (applauncher_settings.themename)
          g_free (applauncher_settings.themename);
      applauncher_settings.themename = g_strdup(newtheme);
      set_gpe_theme (newtheme);
      applications_update_theme (applauncher_settings.themename);
      softkeys_update_theme (newtheme);
      homescreen_update_theme (newtheme);
      g_message ("setting theme");
    }
}

/* Background colour change event handler. */
static void
update_background_colour (GdkColor colour)
{
  homescreen_update_bg_colour (colour);
}

/* Label colour change event handler. */
static void
update_label_colour (const gchar* colour)
{
  homescreen_update_label_colour (colour);
  softkeys_update_label_colour (colour);
}

/* Status label colour change event handler. */
static void
update_status_colour (const gchar* colour)
{
  homescreen_update_status_label_colour (colour);
}

/* Icon size change change event handler. */
static void
update_icon_size (gint size)
{
  applications_update_itemsize (size);
}

/* Layout change event handler. */
static void
update_item_layout (gint type)
{
  applications_update_item_layout (type);
}

/* Wallpaper change event handler. */
static void
update_background_pixmap (const gchar* filename)
{
  homescreen_update_bg_pixmap (filename);
}

/* Main Xsettings event callback. 
 * Check if we are interested in the changed setting and if a setting is 
 * recognized track it in the settings mask and call the event handler. */
static void
notify_func (const char *name, XSettingsAction action,
             XSettingsSetting *setting, void *cb_data)
{
	g_message ("xsetting notification received");
  /* applauncher specific settings */
  if (strncmp (name, KEY_BASE, strlen (KEY_BASE)) == 0)
    {
      char *p = (char*)name + strlen (KEY_BASE);

      if (!strcmp (p, "layout") && setting->type == XSETTINGS_TYPE_INT)
        {
          gint type = setting->data.v_int;
          if ((type < 0) || (type >= LAYOUT_MAX)) 
              type = LAYOUT_STACK;
          applauncher_settings.item_layout = type;
          applauncher_settings.xset_mask |= XSET_ITEM_LAYOUT;

          update_item_layout (type);
        }
      if (!strcmp (p, "bgcol") && setting->type == XSETTINGS_TYPE_STRING)
        {
          GdkColor c;
          if (gdk_color_parse (setting->data.v_string, &c))
            {
              if (applauncher_settings.bg_colour)
                g_free(applauncher_settings.bg_colour);
              applauncher_settings.bg_colour = g_strdup (setting->data.v_string);
              applauncher_settings.xset_mask |= XSET_BG_COLOUR;
              update_background_colour (c);
            }
        }
      if (!strcmp (p, "labelcol") && setting->type == XSETTINGS_TYPE_STRING)
        {
          GdkColor c;
          if (gdk_color_parse (setting->data.v_string, &c))
            {
              if (applauncher_settings.home_font_colour)
                g_free(applauncher_settings.home_font_colour);
              applauncher_settings.home_font_colour = g_strdup (setting->data.v_string);
              applauncher_settings.xset_mask |= XSET_HOME_FONT_COLOUR;
              /* We use the string representation here because labels take 
                 it in this form instead of GdkColor. */
              update_label_colour (applauncher_settings.home_font_colour);
            }
        }
      if (!strcmp (p, "statuscol") && setting->type == XSETTINGS_TYPE_STRING)
        {
          GdkColor c;
          if (gdk_color_parse (setting->data.v_string, &c))
            {
              update_status_colour (setting->data.v_string);
            }
        }
    }
  if (strcasecmp (name, KEY_BASE "/wallpaper") == 0)
    {
      if (setting && setting->type == XSETTINGS_TYPE_STRING)
        {
          enum image_type type = SCALED;
          gchar *path = setting->data.v_string;
          if (!strncmp (path, "img-tiled:", 10))
            {
              path += 10;
              type = TILED;
            }
          else if (!strncmp (path, "img-stretched:", 14))
            {
              path += 14;
              type = SCALED;
            }
          else if (!strncmp (path, "img-centered:", 13))
            {
              path += 13;
              type = CENTERED;
            }
          if (applauncher_settings.bg_filename)
            g_free(applauncher_settings.bg_filename);
          applauncher_settings.bg_filename = g_strdup (path);
          applauncher_settings.xset_mask |= XSET_BG_FILENAME;

          update_background_pixmap (path);
        }
      else if (setting == NULL)
        {
          if (applauncher_settings.bg_filename)
            g_free(applauncher_settings.bg_filename);
          applauncher_settings.bg_filename = NULL;
          
          update_background_pixmap (NULL);
        }
    }
  else if (strcasecmp (name, KEY_BASE "/iconsize") == 0)
    {
      if (setting && setting->type == XSETTINGS_TYPE_INT)
        {
          gint size = setting->data.v_int;
          if (size < 12)
              size = 12;
          if (size > 48)
              size = 48;
          applauncher_settings.item_size = size;
          applauncher_settings.xset_mask |= XSET_ITEM_SIZE;

          update_icon_size (size);
        }
    }
  else if (strncmp (name, KEY_THEME, strlen (KEY_THEME)) == 0)
    {
      char *p = (char*)name + strlen (KEY_THEME);

      if (!strcmp (p, "ThemeName") && setting->type == XSETTINGS_TYPE_STRING)
        {
          if (setting->data.v_string && strlen (setting->data.v_string))
              update_theme (setting->data.v_string);
        }
    } 
}

/* Xsettings event filter function. */
static GdkFilterReturn
xsettings_event_filter (GdkXEvent *xevp, GdkEvent *ev, gpointer p)
{
  if (xsettings_client_process_event (client, (XEvent *)xevp))
    return GDK_FILTER_REMOVE;

  return GDK_FILTER_CONTINUE;
}

/* Xsettings watch function - ususally no need to touch this. */
static void 
watch_func (Window window, Bool is_start, long mask, void  *cb_data)
{
  GdkWindow *gdkwin;
  
  gdkwin = gdk_window_lookup (window);

  if (is_start)
    {
      if (!gdkwin)
        gdkwin = gdk_window_foreign_new (window);
      else
        g_object_ref (gdkwin);

      gdk_window_add_filter (gdkwin, xsettings_event_filter, NULL);
    }
  else
    {
      assert (gdkwin);
      g_object_unref (gdkwin);
      gdk_window_remove_filter (gdkwin, xsettings_event_filter, NULL);
    }
}

/* Xsettings configuration initialisation function. */
gboolean
applauncher_start_xsettings (void)
{
  Display *dpy = GDK_DISPLAY ();

  client = xsettings_client_new (dpy, DefaultScreen (dpy), notify_func, watch_func, NULL);
  if (client == NULL)
    {
      g_printerr ("Cannot create XSettings client\n");
      return FALSE;
    }

  return TRUE;
}
